import tempfile
import os

from cap_anndata.reader import read_h5ad, read_directly
from test.context import get_base_anndata
import pytest
import warnings


def prepare_h5ad_file(name):
    adata = get_base_anndata()
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, name)
    adata.write_h5ad(file_path)
    return file_path


@pytest.mark.parametrize("edit", [True, False])
def test_read_in_context(edit):
    file_path = prepare_h5ad_file("test_read_in_context.h5ad")

    with read_h5ad(file_path=file_path, edit=edit) as cap_adata:
        assert cap_adata is not None, "CapAnnData file must be valid!"
        cap_adata.read_obs()
        cap_adata.read_uns()
        if edit:
            cap_adata.overwrite()

    os.remove(file_path)


@pytest.mark.parametrize("edit", [True, False])
def test_read_as_function(edit):
    file_path = prepare_h5ad_file("test_read_as_function.h5ad")

    cap_adata = read_h5ad(file_path=file_path, edit=edit)
    assert cap_adata is not None, "CapAnnData file must be valid!"
    cap_adata.read_obs()
    cap_adata.read_uns()
    if edit:
        cap_adata.overwrite()

    cap_adata.file.close()
    os.remove(file_path)


@pytest.mark.parametrize("edit", [True, False])
def test_read_directly(edit):
    # TODO: remove deprecated function and unit test
    file_path = prepare_h5ad_file("test_read_directly.h5ad")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        cap_adata = read_directly(file_path=file_path, edit=edit)

    assert cap_adata is not None, "CapAnnData file must be valid!"
    cap_adata.read_obs()
    cap_adata.read_uns()
    if edit:
        cap_adata.overwrite()

    cap_adata.file.close()
    os.remove(file_path)



import pandas as pd
import numpy as np
from typing import List, Any, Union

from pandas._typing import Self
from pandas.core.generic import bool_t


class CapAnnDataDF(pd.DataFrame):
    """
    The class to expand the pandas DataFrame behaviour to support partial
    reading and writing of AnnData obs and var (raw.var) fields.
    The main feature of the class is handling <column-order> attribute
    which must be a copy of h5py.Group attribute
    """

    _metadata = ["column_order"]

    def column_order_array(self) -> np.array:
        order = self.column_order
        if order is not None and isinstance(order, List):
            # Convert it to numpy array of str elements
            return np.array(order, dtype=object)
        else:
            return order

    def rename_column(self, old_name: str, new_name: str) -> None:
        i = np.where(self.column_order_array() == old_name)[0]
        tmp_array = self.column_order_array().copy()
        tmp_array[i] = new_name
        self.column_order = tmp_array.copy()
        self.rename(columns={old_name: new_name}, inplace=True)

    def remove_column(self, col_name: str) -> None:
        i = np.where(self.column_order_array() == col_name)[0]
        self.column_order = np.delete(self.column_order_array(), i)
        self.drop(columns=[col_name], inplace=True)

    def __setitem__(self, key, value) -> None:
        if key not in self.column_order_array():
            self.column_order = np.append(self.column_order_array(), key)
        return super().__setitem__(key, value)

    @classmethod
    def from_df(cls, df: pd.DataFrame, column_order: Union[np.array, List[str], None] = None) -> Self:
        if column_order is None:
            column_order = df.columns.to_numpy()
        elif isinstance(column_order, List):
            column_order = np.array(column_order)
        new_inst = cls(df)
        new_inst.column_order = column_order
        return new_inst

    def join(self, other: Any, **kwargs) -> Self:
        result = super().join(other=other, **kwargs)
        if isinstance(other, CapAnnDataDF):
            new_columns = [
                col for col in other.column_order_array() if col not in self.column_order_array()
            ]
        else:
            new_columns = [col for col in other.columns if col not in self.column_order_array()]
        column_order = np.append(self.column_order_array(), new_columns)
        df = self.from_df(result, column_order=column_order)
        return df

    def merge(self, right, **kwargs) -> Self:
        result = super().merge(right=right, **kwargs)
        if isinstance(right, CapAnnDataDF):
            new_columns = [
                col for col in right.column_order_array() if col not in self.column_order_array()
            ]
        else:
            new_columns = [col for col in right.columns if col not in self.column_order_array()]
        column_order = np.append(self.column_order_array(), new_columns)
        df = self.from_df(result, column_order=column_order)
        return df

    def copy(self, deep: Union[bool_t, None] = True) -> Self:
        column_order = self.column_order_array()
        df = self.from_df(super().copy(deep=deep), column_order=column_order)
        return df



from typing import Set, Any


class CapAnnDataDict(dict):
    __keys_to_remove: Set[str] = None

    def __delitem__(self, __key: Any) -> None:
        self.keys_to_remove.add(__key)
        return super().__delitem__(__key)

    def __setitem__(self, __key: Any, __value: Any) -> None:
        if __value is not None:
            if __key in self.keys_to_remove:
                self.keys_to_remove.remove(__key)
        else:
            self.keys_to_remove.add(__key)
        return super().__setitem__(__key, __value)

    @property
    def keys_to_remove(self) -> Set[str]:
        if self.__keys_to_remove is None:
            self.__keys_to_remove = set()
        return self.__keys_to_remove

    def pop(self, __key: Any, __default: Any = None) -> Any:
        if __key in self:
            self.keys_to_remove.add(__key)
        return super().pop(__key, __default)

    def popitem(self) -> Any:
        item = super().popitem()
        key = item[0]
        self.keys_to_remove.add(key)
        return item



import pandas as pd
import numpy as np
from cap_anndata import CapAnnDataDF


def test_from_df():
    data = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    cap_anndata_df = CapAnnDataDF.from_df(data)

    assert np.allclose(data.values, cap_anndata_df.values)
    assert all(data.columns == cap_anndata_df.column_order)


def test_create_column():
    data = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    cap_anndata_df = CapAnnDataDF.from_df(data, column_order=["A", "B", "D"])
    cap_anndata_df["C"] = [7, 8, 9]

    assert "C" in cap_anndata_df.columns
    assert "C" in cap_anndata_df.column_order


def test_rename_column():
    data = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    cap_anndata_df = CapAnnDataDF.from_df(data)
    cap_anndata_df.rename_column("A", "A_renamed")

    assert "A_renamed" in cap_anndata_df.columns
    assert "A" not in cap_anndata_df.columns
    assert "A_renamed" in cap_anndata_df.column_order


def test_remove_column():
    data = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    cap_anndata_df = CapAnnDataDF.from_df(data)
    cap_anndata_df.remove_column("B")

    assert "B" not in cap_anndata_df.columns
    assert "B" not in cap_anndata_df.column_order


def test_from_df_class_method():
    data = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    new_df = CapAnnDataDF.from_df(data)
    assert list(new_df.column_order) == ["A", "B"]


def test_column_order_integrity():
    data = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    cap_anndata_df = CapAnnDataDF.from_df(data)
    cap_anndata_df["C"] = [7, 8, 9]
    cap_anndata_df.rename_column("A", "A_renamed")
    cap_anndata_df.remove_column("B")

    expected_order = ["A_renamed", "C"]
    assert list(cap_anndata_df.column_order) == expected_order


def test_join():
    data1 = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    data2 = pd.DataFrame({"C": [7, 8, 9], "D": [10, 11, 12]})
    cap_anndata_df1 = CapAnnDataDF.from_df(data1)
    cap_anndata_df2 = cap_anndata_df1.join(data2, how="left")

    expected_order = ["A", "B", "C", "D"]
    assert list(cap_anndata_df2.column_order) == expected_order
    assert cap_anndata_df2.shape == (3, 4)


def test_merge():
    data1 = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    data2 = pd.DataFrame({"A": [2, 3, 4], "D": [10, 11, 12]})
    cap_anndata_df1 = CapAnnDataDF.from_df(data1)

    cap_anndata_df1 = cap_anndata_df1.merge(data2, how="inner", on="A")

    expected_order = ["A", "B", "D"]
    assert list(cap_anndata_df1.column_order) == expected_order
    assert cap_anndata_df1.shape == (2, 3)



import anndata as ad
from scipy.sparse import csr_matrix, csc_matrix
import numpy as np
import tempfile
import os
import pandas as pd
import pytest

from cap_anndata import CapAnnDataDF
from test.context import get_base_anndata
from cap_anndata.reader import read_h5ad


def get_filled_anndata(n_rows: int = 10, n_genes: int = 10, sparse=False) -> ad.AnnData:
    adata = get_base_anndata(n_rows, n_genes, sparse)

    adata.obs["cell_type"] = [f"cell_{i%3}" for i in range(adata.shape[0])]
    adata.obs["number"] = [i / 10 for i in range(adata.shape[0])]
    adata.obs.index = [f"obs_{i}" for i in range(adata.shape[0])]

    adata.var.index = [f"gene_{i}" for i in range(adata.shape[1])]
    adata.var["filtered"] = [i > 4 for i in range(adata.shape[1])]
    adata.var["gene_names"] = [f"gene_name_{i}" for i in range(adata.shape[1])]
    adata.var["dispersion"] = [i / 100 for i in range(adata.shape[1])]

    adata.raw = adata
    return adata


def save_filled_anndata(
    file_name: str, n_rows: int = 10, n_genes: int = 10, sparse=False
) -> str:
    adata = get_filled_anndata(n_rows, n_genes, sparse)
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, file_name)
    adata.write_h5ad(file_path)
    return file_path


def test_read_shape():
    n_rows = 10
    n_genes = 20
    adata = get_base_anndata(n_rows, n_genes)
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_read_shape.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path) as cap_adata:
        shape = cap_adata.shape

    os.remove(file_path)

    assert shape == (n_rows, n_genes), "Shape axis size is incorrect!"

    for i in [0, 1]:
        assert type(shape[i]) == int, "Shape axis type is wrong!"


def test_read_df():
    adata = get_filled_anndata()
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_read_obs.h5ad")

    adata.write_h5ad(file_path)

    with read_h5ad(file_path) as cap_adata:
        cap_adata.read_obs()
        cap_adata.read_var()
        cap_adata.raw.read_var()

    os.remove(file_path)
    pd.testing.assert_frame_equal(adata.obs, cap_adata.obs, check_frame_type=False)
    pd.testing.assert_frame_equal(adata.var, cap_adata.var, check_frame_type=False)
    pd.testing.assert_frame_equal(
        adata.raw.var, cap_adata.raw.var, check_frame_type=False
    )


def test_partial_read():
    adata = get_filled_anndata()
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_partial_read.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path) as cap_adata:
        cap_adata.read_obs(columns=["cell_type"])
        cap_adata.read_obs(columns=["cell_type"])
        cap_adata.read_var(columns=["dispersion"])
        cap_adata.raw.read_var(columns=["dispersion"])

    os.remove(file_path)

    assert len(adata.obs.columns) == len(cap_adata.obs.column_order)
    assert len(adata.var.columns) == len(cap_adata.var.column_order)
    assert len(adata.raw.var.columns) == len(cap_adata.raw.var.column_order)

    assert len(cap_adata.obs.columns) == 1
    assert len(cap_adata.var.columns) == 1
    assert len(cap_adata.raw.var.columns) == 1

    pd.testing.assert_index_equal(adata.obs.index, cap_adata.obs.index)
    pd.testing.assert_index_equal(adata.var.index, cap_adata.var.index)
    pd.testing.assert_index_equal(adata.raw.var.index, cap_adata.raw.var.index)


def test_overwrite_dataframe_before_read_obs():
    path = "tmp.h5ad"
    x = np.ones((10, 10), dtype=np.float32)
    adata = ad.AnnData(X=x)
    adata.obs["columns"] = "value"
    adata.write_h5ad(path)
    del adata

    with read_h5ad(path, True) as adata:
        # https://github.com/cellannotation/cap-anndata/issues/33
        adata.obs["new_column"] = "new_value"
        adata.overwrite(["obs"])

    with read_h5ad(path) as adata:
        adata.read_obs("new_column")
        assert (adata.obs["new_column"] == "new_value").all(), "Wrong values in column!"

    os.remove(path)


@pytest.mark.parametrize("compression", ["gzip", "lzf"])
def test_overwrite_df(compression):
    adata = get_filled_anndata()
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_overwrite_df.h5ad")
    adata.write_h5ad(file_path)

    new_obs_index = None
    with read_h5ad(file_path, edit=True) as cap_adata:
        # Modify 'obs'
        cap_adata.read_obs(columns=["cell_type"])
        cap_adata.obs["cell_type"] = [
            f"new_cell_type_{i%2}" for i in range(cap_adata.shape[0])
        ]
        cap_adata.obs["const_str"] = "some string"
        # Modify obs 'index'
        new_obs_index = [s + "_new" for s in cap_adata.obs.index]
        cap_adata.obs.index = new_obs_index
        ref_obs = cap_adata.obs.copy()

        # Modify 'var'
        cap_adata.read_var()
        cap_adata.var["gene_names"] = [
            f"new_gene_{i}" for i in range(cap_adata.shape[1])
        ]
        cap_adata.var["extra_info"] = np.random.rand(cap_adata.shape[1])
        ref_var = cap_adata.var.copy()

        # Modify 'raw.var', assuming 'raw' is also a CapAnnData
        cap_adata.raw.read_var()
        cap_adata.raw.var["gene_names"] = [
            f"raw_new_gene_{i}" for i in range(cap_adata.raw.shape[1])
        ]
        cap_adata.raw.var["extra_info"] = np.random.rand(cap_adata.shape[1])
        ref_raw_var = cap_adata.raw.var.copy()

        cap_adata.overwrite(["obs", "var", "raw.var"], compression=compression)

    adata = ad.read_h5ad(file_path)
    os.remove(file_path)

    # Assert changes in 'obs'
    assert all([c in adata.obs.columns for c in ref_obs.columns])
    pd.testing.assert_frame_equal(
        ref_obs, adata.obs[ref_obs.columns.to_list()], check_frame_type=False
    )
    assert (adata.obs.index == new_obs_index).all(), "Index must be changed!"

    # Assert changes in 'var'
    assert all([c in adata.var.columns for c in ref_var.columns])
    pd.testing.assert_frame_equal(
        ref_var, adata.var[ref_var.columns.to_list()], check_frame_type=False
    )

    # Assert changes in 'raw.var'
    assert all([c in adata.raw.var.columns for c in ref_raw_var.columns])
    pd.testing.assert_frame_equal(
        ref_raw_var,
        adata.raw.var[ref_raw_var.columns.to_list()],
        check_frame_type=False,
    )


@pytest.mark.parametrize("sparse", [False, True])
@pytest.mark.parametrize("vertical_slice", [None, False, True, "mask"])
def test_link_x(sparse, vertical_slice):
    adata = get_filled_anndata(sparse=sparse)
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_link_x.h5ad")
    adata.write_h5ad(file_path)

    if vertical_slice is None:
        s_ = np.s_[:]
    elif vertical_slice == "mask":
        mask = np.array([i < 5 for i in range(adata.shape[0])])
        s_ = np.s_[mask, :5]
    else:
        # slice over var or obs
        s_ = np.s_[:, 0:5] if vertical_slice else np.s_[0:5, :]

    with read_h5ad(file_path) as cap_adata:
        x = cap_adata.X[s_]
        raw_x = cap_adata.raw.X[s_]

    os.remove(file_path)
    if sparse:
        assert np.allclose(adata.X.toarray()[s_], x.toarray())
        assert np.allclose(adata.raw.X.toarray()[s_], raw_x.toarray())
    else:
        assert np.allclose(adata.X[s_], x)
        assert np.allclose(adata.raw.X[s_], raw_x)


@pytest.mark.parametrize("sparse", [False, True])
def test_shape(sparse):
    n_rows = 15
    n_genes = 25

    adata = get_filled_anndata(n_rows, n_genes, sparse)
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_shape.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path) as cap_adata:
        shape = cap_adata.shape
        shape_raw = cap_adata.raw.shape

    os.remove(file_path)
    for sh in [shape, shape_raw]:
        assert sh == (n_rows, n_genes)


def test_read_obsm_varm():
    adata = get_filled_anndata()
    emb_names = [f"X_test{i}" for i in range(2)]

    for emb in emb_names:
        adata.obsm[emb] = np.random.random(size=(adata.shape[0], 2))
        adata.varm[emb] = np.random.random(size=(adata.shape[1], 2))

    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_read_obsm.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path) as cap_adata:
        for emb in emb_names:
            assert emb in cap_adata.obsm_keys()
            assert cap_adata.obsm[emb].shape == adata.obsm[emb].shape
            assert emb in cap_adata.varm.keys()
            assert cap_adata.varm[emb].shape == adata.varm[emb].shape
            assert np.allclose(adata.obsm[emb], cap_adata.obsm[emb][:])
            assert np.allclose(adata.varm[emb], cap_adata.varm[emb][:])

    os.remove(file_path)


def test_create_remove_obsm():
    shape = (10, 10)
    file_path = save_filled_anndata(
        "test_create_remove_obsm.h5ad", n_rows=shape[0], n_genes=shape[1]
    )
    emb_name = "X_test"
    emb = np.random.random(size=(shape[0], 2))

    with read_h5ad(file_path, edit=True) as cap_adata:
        cap_adata.create_obsm(name=emb_name, matrix=emb)
        cap_adata.create_obsm(name=emb_name + "_pop", matrix=emb)

    with read_h5ad(file_path) as cap_adata:
        assert emb_name in cap_adata.obsm.keys()
        assert np.allclose(emb, cap_adata.obsm[emb_name][:])
        assert emb_name + "_pop" in cap_adata.obsm.keys()
        assert np.allclose(emb, cap_adata.obsm[emb_name + "_pop"][:])

    with read_h5ad(file_path, edit=True) as cap_adata:
        cap_adata.remove_obsm(emb_name)
        cap_adata.obsm.pop(emb_name + "_pop")
        cap_adata.overwrite(["obsm"])

    with read_h5ad(file_path) as cap_adata:
        assert emb_name not in cap_adata.obsm.keys()
        assert emb_name + "_pop" not in cap_adata.obsm.keys()


def test_create_remove_varm():
    shape = (10, 10)
    file_path = save_filled_anndata(
        "test_create_varm.h5ad", n_rows=shape[0], n_genes=shape[1]
    )
    emb_name = "X_test"
    emb = np.random.random(size=(shape[1], 2))

    with read_h5ad(file_path, edit=True) as cap_adata:
        cap_adata.create_varm(name=emb_name, matrix=emb)
        cap_adata.create_varm(name=emb_name + "_pop", matrix=emb)

    with read_h5ad(file_path) as cap_adata:
        assert emb_name in cap_adata.varm.keys()
        assert np.allclose(emb, cap_adata.varm[emb_name][:])
        assert emb_name + "_pop" in cap_adata.varm.keys()
        assert np.allclose(emb, cap_adata.varm[emb_name + "_pop"][:])

    with read_h5ad(file_path, edit=True) as cap_adata:
        cap_adata.remove_varm(emb_name)
        cap_adata.varm.pop(emb_name + "_pop")
        cap_adata.overwrite(["varm"])

    with read_h5ad(file_path) as cap_adata:
        assert emb_name not in cap_adata.varm.keys()
        assert emb_name + "_pop" not in cap_adata.varm.keys()


def test_read_uns():
    adata = get_base_anndata()
    key1, key2 = "key1", "key2"
    keys = (key1, key2)

    adata.uns = {k: {k: k} for k in keys}
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_read_uns.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path) as cap_adata:
        for k in keys:
            assert k in cap_adata.uns

        cap_adata.read_uns(keys=[key1])

        assert cap_adata.uns[key1] == adata.uns[key1]  # connected
        assert cap_adata.uns[key2] != adata.uns[key2]  # not connected

    os.remove(file_path)


@pytest.mark.parametrize("compression", ["gzip", "lzf"])
def test_modify_uns(compression):
    adata = get_base_anndata()
    adata.uns = {
        "field_to_ignore": list(range(100)),
        "field_to_rename": "value",
        "field_to_expand": {"key1": {}},
        "field_to_modify": {"a": "b"},
    }
    new_name = "renamed_field"
    d_to_exp = {"sub_key1": "v1", "sub_key2": "v2"}
    v_to_mod = "value"

    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_modify_uns.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path, edit=True) as cap_adata:
        cap_adata.read_uns(
            keys=["field_to_rename", "field_to_expand", "field_to_modify"]
        )

        cap_adata.uns[new_name] = cap_adata.uns.pop("field_to_rename")
        cap_adata.uns["field_to_expand"]["key1"] = d_to_exp
        cap_adata.uns["field_to_modify"] = v_to_mod

        cap_adata.overwrite(["uns"], compression=compression)

    adata = ad.read_h5ad(file_path)

    assert adata.uns is not None
    assert len(adata.uns.keys()) == 4
    assert new_name in adata.uns.keys()
    assert adata.uns["field_to_expand"]["key1"] == d_to_exp
    assert adata.uns["field_to_modify"] == v_to_mod


def test_empty_obs_override():
    """
    especially for solving the issue:
    https://github.com/cellannotation/cap-anndata/pull/5
    """
    adata = get_base_anndata()
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_modify_uns.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path, edit=True) as cap_adata:
        cap_adata.read_obs()

        cap_adata.obs["cell_type_1"] = pd.Series(
            data=np.nan, index=cap_adata.obs.index, dtype="category"
        )
        cap_adata.obs["cell_type_new"] = pd.Series(
            data=np.nan, index=cap_adata.obs.index, dtype="category"
        )
        cap_adata.overwrite(fields=["obs"])


def test_obs_var_keys():
    adata = get_filled_anndata()
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_obs_keys.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path) as cap_adata:
        obs_keys_before_read = cap_adata.obs_keys()
        var_keys_before_read = cap_adata.var_keys()
        cap_adata.read_obs()
        cap_adata.read_var()
        obs_keys_after_read = cap_adata.obs_keys()
        var_keys_after_read = cap_adata.var_keys()

    assert obs_keys_before_read == adata.obs_keys()
    assert obs_keys_after_read == adata.obs_keys()
    assert var_keys_before_read == adata.var_keys()
    assert var_keys_after_read == adata.var_keys()
    os.remove(file_path)


def test_reset_read():
    adata = get_filled_anndata()
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_flush_read.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path) as cap_adata:
        cap_adata.read_obs(columns=["cell_type"])
        cap_adata.obs["cell_type"] = list(range(cap_adata.shape[0]))

        cap_adata.read_obs(columns=["number"])
        assert cap_adata.obs.columns.size == 2
        assert all(cap_adata.obs.cell_type.values == list(range(cap_adata.shape[0])))

        cap_adata.read_obs(columns=["cell_type"])
        pd.testing.assert_series_equal(cap_adata.obs.cell_type, adata.obs.cell_type)

        cap_adata.read_obs(columns=["number"], reset=True)
        assert cap_adata.obs.columns.size == 1
    os.remove(file_path)


def test_obs_last_column_removal():
    col_name = "cell_type"
    adata = ad.AnnData(X=np.ones(shape=(10, 10), dtype=np.float32))
    adata.obs[col_name] = col_name

    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_obs_last_column_removal.h5ad")
    adata.write(filename=file_path)

    # Remove last column
    with read_h5ad(file_path, edit=True) as cap_adata:
        cap_adata.read_obs()
        cap_adata.obs.remove_column(col_name=col_name)
        cap_adata.overwrite(["obs"])

    # Check no any issues on read with updated file
    with read_h5ad(file_path) as cap_adata:
        cap_adata.read_obs()
        assert col_name not in cap_adata.obs.columns

    # Check compatability with anndata
    adata = ad.read_h5ad(file_path)
    os.remove(file_path)


@pytest.mark.parametrize("field", ["obs", "var", "raw.var"])
def test_df_setter(field):
    def set_field(field, new_df, cap_adata):
        if field == "obs":
            cap_adata.obs = new_df
        elif field == "var":
            cap_adata.var = new_df
        else:
            cap_adata.raw.var = new_df

    n_rows, n_genes = 10, 5
    adata = get_filled_anndata(n_rows=n_rows, n_genes=n_genes)
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_obs_setter.h5ad")
    adata.write(filename=file_path)

    new_df = pd.DataFrame(
        index=range(n_rows if field == "obs" else n_genes), columns=["new_column"]
    )

    with read_h5ad(file_path, edit=True) as cap_adata:
        cap_adata.read_obs()
        cap_adata.read_var()
        cap_adata.raw.read_var()

        # test bad format
        try:
            set_field(field, new_df, cap_adata)
        except TypeError:
            pass
        except Exception as e:
            assert False, f"Unexpected exception: {e}"
        else:
            assert False, "Expected TypeError"

        # test good format
        new_df = CapAnnDataDF.from_df(new_df)
        set_field(field, new_df, cap_adata)

        # test bad shape
        new_df = CapAnnDataDF.from_df(new_df[: new_df.shape[0] // 2])

        try:
            set_field(field, new_df, cap_adata)
        except ValueError:
            pass
        except Exception as e:
            assert False, f"Unexpected exception: {e}"
        else:
            assert False, "Expected ValueError"


def test_layer_base_operations():
    adata = get_filled_anndata()
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_layers.h5ad")

    adata.write_h5ad(file_path)

    layer_name_dense = "layer_dense"
    layer_name_sparse = "layer_sparse"

    shape = (10, 10)
    data = np.random.random(shape)
    dense_array = data
    sparse_array = csr_matrix(data)

    # Test add layers
    with read_h5ad(file_path, edit=True) as cap_adata:
        cap_adata.create_layer(name=layer_name_dense, matrix=dense_array)
        cap_adata.create_layer(name=layer_name_sparse, matrix=sparse_array)

    with read_h5ad(file_path) as cap_adata:
        assert np.array_equal(
            dense_array, cap_adata.layers[layer_name_dense]
        ), "Must be correct dense matrix!"
        assert np.array_equal(
            sparse_array.todense(), cap_adata.layers[layer_name_sparse][:].todense()
        ), "Must be correct sparse matrix!"

    # Test remove layers
    with read_h5ad(file_path, edit=True) as cap_adata:
        cap_adata.remove_layer(layer_name_dense)
        cap_adata.remove_layer(layer_name_sparse)

    with read_h5ad(file_path) as cap_adata:
        assert (
            layer_name_dense not in cap_adata.layers.keys()
        ), "Dense matrix must be removed!"
        assert (
            layer_name_sparse not in cap_adata.layers.keys()
        ), "Sparse matrix must be removed!"

    # Test modify layers
    layer_name_edit = "layer_for_edit"
    data = np.ones(shape)
    with read_h5ad(file_path, edit=True) as cap_adata:  # fill layer
        cap_adata.create_layer(name=layer_name_edit, matrix=data)
    with read_h5ad(file_path, edit=True) as cap_adata:  # modify backed
        cap_adata.layers[layer_name_edit][0:1, 0:1] = 0
    with read_h5ad(file_path) as cap_adata:  # check is changed
        assert False == np.array_equal(
            data, cap_adata.layers[layer_name_edit][:]
        ), "Layer matrix must be edited previously!"

    os.remove(file_path)


def test_layer_create_append():
    adata = get_filled_anndata()
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_layers.h5ad")

    adata.write_h5ad(file_path)

    shape = (10, 10)
    data = np.random.random(shape)

    # Test add empty dense and sparse data layers and edit them
    layer_name_empty_dense = "layer_empty_dense"

    def layer_name_empty_sparse(format: str):
        return f"layer_empty_sparse_{format}"

    def sparse_array_class(format: str):
        return csr_matrix if format == "csr" else csc_matrix

    with read_h5ad(file_path, edit=True) as cap_adata:
        cap_adata.create_layer(
            name=layer_name_empty_dense,
            matrix=None,
            matrix_shape=shape,
            data_dtype=np.float32,
            format="dense",
        )
        for format in ["csr", "csc"]:
            array_class = sparse_array_class(format)
            sparse_array = array_class(data)
            name = layer_name_empty_sparse(format)
            cap_adata.create_layer(
                name=name,
                matrix=None,
                matrix_shape=shape,
                data_dtype=sparse_array.data.dtype,
                format=format,
            )
    sparse_info = {}
    with read_h5ad(file_path, edit=True) as cap_adata:
        # Modify dense dataset
        cap_adata.layers[layer_name_empty_dense][0, 0] = 1
        # Modify sparse datasets
        for format in ["csr", "csc"]:
            name = layer_name_empty_sparse(format)
            sparse_dataset = cap_adata.layers[name]
            chunk_shape = (1, 10) if format == "csr" else (10, 1)
            array_class = sparse_array_class(format)
            chunk_data = array_class(np.ones(chunk_shape))
            sparse_dataset.append(chunk_data)
            sparse_info[format] = {
                "shape": chunk_data.data.shape,
                "dtype": chunk_data.data.dtype,
            }
    with read_h5ad(file_path) as cap_adata:  # check is changed
        assert np.any(
            cap_adata.layers[layer_name_empty_dense][:] == 1
        ), "Dense layer is not changed!"
        for format in ["csr", "csc"]:
            name = layer_name_empty_sparse(format)
            matrix = cap_adata.layers[name][:]
            assert np.any(
                matrix.toarray() == 1
            ), f"Layer {format} matrix must be edited previously!"
            assert (
                matrix.data.shape == sparse_info[format]["shape"]
            ), "shape is incorrect!"
            assert matrix.data.dtype == sparse_info[format]["dtype"], "dtype is wrong!"

    os.remove(file_path)


def test_read_obsp_varp():
    shape = (10, 10, 2)
    adata = ad.AnnData(X=np.ones(shape[:2], dtype=np.float32))
    rng = np.random.default_rng()
    adata.obsp["obsp_test"] = rng.random(shape)
    adata.varp["varp_test"] = rng.random(shape)

    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_read_obsp_varp.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path) as cap_adata:
        assert np.allclose(adata.obsp["obsp_test"], cap_adata.obsp["obsp_test"][:])
        assert np.allclose(adata.varp["varp_test"], cap_adata.varp["varp_test"][:])
    os.remove(file_path)


@pytest.mark.parametrize("field", ["obsp", "varp"])
def test_modify_obsp_varp(field):
    shape = (10, 10, 2)
    adata = ad.AnnData(X=np.ones(shape[:2], dtype=np.float32))
    rng = np.random.default_rng()
    name = f"{field}_test"
    getattr(adata, field)[name] = rng.random(shape)
    ada_p = getattr(adata, field)[name]
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, f"test_modify_{field}.h5ad")
    adata.write_h5ad(file_path)

    s1 = np.s_[: shape[0] // 2]
    s2 = np.s_[shape[0] // 2 :]

    with read_h5ad(file_path, edit=True) as cap_adata:
        f = getattr(cap_adata, field)[name]
        f.write_direct(np.zeros_like(ada_p), s1, s1)

    with read_h5ad(file_path, edit=False) as cap_adata:
        cap_p = getattr(cap_adata, field)[name]
        assert (cap_p[s1] == 0).all()
        assert np.allclose(ada_p[s2], cap_p[s2])

    # test add new obsp/varp
    new_matrix = rng.random(shape)
    new_name = f"{field}_new"
    with read_h5ad(file_path, edit=True) as cap_adata:
        f = getattr(cap_adata, f"create_{field}")
        f(new_name, matrix=new_matrix)
        assert new_name in getattr(cap_adata, field).keys()

    with read_h5ad(file_path, edit=False) as cap_adata:
        assert np.allclose(new_matrix, getattr(cap_adata, field)[new_name][:])

    with read_h5ad(file_path, edit=True) as cap_adata:
        # test directly remove via dedicated method
        f = getattr(cap_adata, f"remove_{field}")
        f(new_name)
        assert new_name not in cap_adata.obsp.keys()

    with read_h5ad(file_path, edit=True) as cap_adata:
        # test removing via pop
        f = getattr(cap_adata, field)
        f.pop(name)
        cap_adata.overwrite([field])

    with read_h5ad(file_path, edit=False) as cap_adata:
        # test that field is removed
        assert new_name not in getattr(cap_adata, field).keys()
        assert name not in getattr(cap_adata, field).keys()
        assert len(getattr(cap_adata, field).keys()) == 0

    os.remove(file_path)


def test_main_var_layers():
    var_index = [f"ind_{i}" for i in range(10)]
    raw_var_index = [f"raw_ind_{i}" for i in range(10)]

    x = np.eye(10, dtype=np.float32)
    raw_x = x * 2
    adata = ad.AnnData(X=raw_x)
    adata.var.index = raw_var_index
    adata.raw = adata
    adata.X = x
    adata.var.index = var_index
    
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_main_var_layers.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path) as cap_anndata:
        assert cap_anndata.var.index.tolist() == var_index
        assert cap_anndata.raw.var.index.tolist() == raw_var_index
        assert np.allclose(cap_anndata.X[:], x)
        assert np.allclose(cap_anndata.raw.X[:], raw_x)

    os.remove(file_path)


@pytest.mark.parametrize("name", ["barcodes", "", None])
def test_modify_index(name):
    adata = get_base_anndata()
    
    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_main_var_layers.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path=file_path, edit=True) as cap_adata:
        cap_adata.read_obs()
        cap_adata.overwrite(["obs"])
    
    cap_adata = ad.read_h5ad(file_path)
    pd.testing.assert_frame_equal(
        left=adata.obs,
        right=cap_adata.obs,
        check_dtype=True,
        check_index_type=True,
        check_names=True,
    )

    with read_h5ad(file_path=file_path, edit=True) as cap_adata:
        cap_adata.read_obs()
        cap_adata.obs.index = pd.Series(data=[f"cell_{i}" for i in range(cap_adata.shape[0])], name=name)
        cap_adata.overwrite(["obs"])
    
    with read_h5ad(file_path=file_path, edit=False) as cap_adata:
        cap_adata.read_obs()
        obs = cap_adata.obs
    
        assert obs is not None, "DataFrame must be loaded!"
        assert obs.index is not None, "DataFrame must have Index!"
        if not name:
            assert obs.index.name == None, "Index name must not be set!"
        else:
            assert obs.index.name == name, "Index name must be set!"
        assert obs.index.to_list() == [f"cell_{i}" for i in range(cap_adata.shape[0])], "Wrong index values!"


def test_column_order_changes():
    adata = get_base_anndata(n_rows = 3, n_genes = 2, sparse=False)

    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_column_order.h5ad")
    adata.write_h5ad(file_path)

    data = {"A": [1, 2, 3], "B": [4, 5, 6]}
    with read_h5ad(file_path=file_path, edit=True) as cap_adata:
        df = pd.DataFrame(data)
        cap_df = CapAnnDataDF.from_df(df)
        cap_adata.obs = CapAnnDataDF.from_df(cap_df)
        cap_adata.overwrite(["obs"])

    new_column_order = list(data.keys())
    new_column_order.reverse()
    with read_h5ad(file_path=file_path, edit=True) as cap_adata:
        cap_adata.read_obs()
        df = cap_adata.obs[new_column_order] # change order via dataframe
        cap_df = CapAnnDataDF.from_df(df)
        cap_adata.obs = cap_df
        cap_adata.overwrite(["obs"])

    new_column_order.reverse()
    with read_h5ad(file_path=file_path, edit=True) as cap_adata:
        cap_adata.read_obs()
        cap_df = cap_adata.obs
        cap_df.column_order = new_column_order # change order via column_order
        cap_adata.obs = cap_df
        cap_adata.overwrite(["obs"])

    with read_h5ad(file_path=file_path) as cap_adata:
        cap_adata.read_obs()
        assert list(cap_adata.obs.column_order) == new_column_order
        assert list(cap_adata.obs.columns) == new_column_order


def test_df_in_obsm():
    adata = get_base_anndata(n_rows = 3, n_genes = 2, sparse=False)
    df = pd.DataFrame(index=adata.obs.index, data={"n": 1})
    adata.obsm["df"] = df

    temp_folder = tempfile.mkdtemp()
    file_path = os.path.join(temp_folder, "test_df_in_obsm.h5ad")
    adata.write_h5ad(file_path)

    with read_h5ad(file_path=file_path, edit=False) as cap_adata:
        assert cap_adata.obsm_keys() == ["df"]
        cap_df = cap_adata.obsm["df"]
    assert cap_df.shape == df.shape
    assert cap_df.columns == df.columns
    assert (cap_df["n"] == df["n"]).all()



from cap_anndata import CapAnnDataDict


def test_basic_init():
    d = {"key1": "value1", "key2": {"value2": "sub_value2"}}
    cap_d = CapAnnDataDict(d)
    assert cap_d == d
    assert len(cap_d.keys_to_remove) == 0
    assert cap_d.get("key1") == "value1"
    assert cap_d.get("key_N") == None


def test_pop():
    cap_d = CapAnnDataDict()
    cap_d["key1"] = "value1"
    cap_d["key2"] = "value2"

    cap_d.pop("key1")

    assert len(cap_d.keys()) == 1
    assert cap_d.keys_to_remove == {"key1"}

    cap_d["key1"] = "new_value"

    assert len(cap_d.keys()) == 2
    assert len(cap_d.keys_to_remove) == 0

    cap_d.popitem()
    assert len(cap_d.keys()) == 1
    assert cap_d.keys_to_remove == {"key1"}

    cap_d["key1"] = {"sk1": "v1", "sk2": "v2"}
    cap_d["key1"].pop("sk2")

    assert len(cap_d.keys()) == 2
    assert len(cap_d.keys_to_remove) == 0



import logging
import anndata as ad
import numpy as np
import h5py
from typing import List, Union, Any, Tuple, Final
import scipy.sparse as ss
from packaging import version

if version.parse(ad.__version__) < version.parse("0.11.0"):
    from anndata.experimental import (
        sparse_dataset,
        read_elem,
        write_elem,
        CSRDataset,
        CSCDataset,
    )
else:
    from anndata.io import (
        sparse_dataset,
        read_elem,
        write_elem,
    )
    from anndata.abc import (
        CSRDataset,
        CSCDataset,
    )

from cap_anndata import CapAnnDataDF, CapAnnDataDict

logger = logging.getLogger(__name__)

X_NOTATION = Union[
    h5py.Dataset, CSRDataset, CSCDataset, None
]
ARRAY_MAPPING_NOTATION = CapAnnDataDict[str, X_NOTATION]
FIELDS_SUPPORTED_TO_OVERWRITE = ["obs", "var", "raw.var", "uns", "layers", "obsm", "varm", "obsp", "varp"]
NotLinkedObject: Final = "__NotLinkedObject"


class BaseLayerMatrixAndDf:
    def __init__(self, file: h5py.File, path_to_content: str = "/") -> None:
        self._file = file
        self._path_to_content = path_to_content
        self._X: X_NOTATION = None

    @property
    def file(self) -> h5py.File:
        return self._file

    @property
    def X(self) -> X_NOTATION:
        if self._X is None:
            self._link_x()
        return self._X

    def _link_x(self) -> None:
        x = self._file[self._path_to_content + "X"]
        if isinstance(x, h5py.Dataset):
            # dense X
            self._X = x
        else:
            # sparse dataset
            self._X = sparse_dataset(x)

    @property
    def shape(self) -> Tuple[int, int]:
        if self.X is not None:
            shape = tuple(map(int, self.X.shape))
        else:
            shape = None
        return shape

    def _lazy_df_load(self, key: str) -> CapAnnDataDF:
        return self._read_df(key=key, columns=[])

    @staticmethod
    def _read_attr(obj: Union[h5py.Group, h5py.Dataset], attr_name: str) -> any:
        attrs = dict(obj.attrs)
        if attr_name not in attrs.keys():
            raise KeyError(f"The {attr_name} doesn't exist!")
        return attrs[attr_name]

    def _read_df(self, key: str, columns: List[str]) -> CapAnnDataDF:
        group_path = self._path_to_content + key
        if group_path not in self._file.keys():
            raise ValueError(f"The group {group_path} doesn't exist in the file!")

        h5_group = self._file[group_path]

        column_order = self._read_attr(h5_group, "column-order")

        if columns is None:
            # read whole df
            df = CapAnnDataDF.from_df(read_elem(h5_group), column_order=column_order)
        else:
            if isinstance(columns, str):
                # single column provided instead of list
                columns = [columns]
            cols_to_read = [c for c in columns if c in column_order]
            df = CapAnnDataDF()
            df.column_order = column_order

            index_col = self._read_attr(h5_group, "_index")
            index = read_elem(h5_group[index_col])
            df.index = index

            for col in cols_to_read:
                df[col] = read_elem(h5_group[col])

        if df.column_order.dtype != object:
            # empty DataFrame will have column_order as float64
            # which leads to failure in overwrite method
            df.column_order = df.column_order.astype(object)
        return df

    def _write_elem(self, dest_key: str, elem: any, compression: str) -> None:
        write_elem(
            self._file, dest_key, elem, dataset_kwargs={"compression": compression}
        )

    def _validate_cap_df(self, cap_df: CapAnnDataDF, axis: int) -> None:
        if not isinstance(cap_df, CapAnnDataDF):
            raise TypeError(
                f"The input should be an instance of CapAnnDataDF class but {type(cap_df)} given!"
            )

        if axis not in [0, 1]:
            raise ValueError("The axis should be either 0 or 1!")

        if cap_df.shape[0] != self.shape[axis]:
            items = "cells" if axis == 0 else "genes"
            raise ValueError(
                f"The number of rows in the input DataFrame should be equal to the number of {items} in the "
                "AnnData object!"
            )

    def _link_array_mapping(self, cap_dict: CapAnnDataDict, key: str) -> None:
        """Method to update given cap_dict with backed array entities from the file."""
        if key not in self._file.keys():
            raise KeyError(f"The key {key} doesn't exist in the file! Ignore linking.")

        group = self._file[key]
        if not isinstance(group, h5py.Group):
            raise ValueError(f"The object {key} must be a group!")

        for entity_name in group.keys():
            entity = group[entity_name]
            if isinstance(entity, h5py.Dataset):
                cap_dict[entity_name] = entity
            elif isinstance(entity, h5py.Group):
                enc_type = dict(entity.attrs).get("encoding-type")
                if enc_type == "dataframe":
                    cap_dict[entity_name] = self._read_df(key="/".join([key, entity_name]), columns=None)
                elif enc_type in ["csc_matrix", "csr_matrix"]:
                    cap_dict[entity_name] = sparse_dataset(entity)
            else:
                raise ValueError(
                    f"Can't link array in {key} due to unsupported type of object: {type(entity)}"
                )

    def _create_new_matrix(
        self,
        dest: str,
        matrix: Union[np.ndarray, ss.csr_matrix, ss.csc_matrix, None] = None,
        matrix_shape: Union[tuple[int, int], None] = None,
        data_dtype: Union[np.dtype, None] = None,
        format: Union[str, None] = None,  # TODO: use Enum instead of str
        compression: str = "lzf",
    ) -> None:
        if matrix is not None:
            self._write_elem(dest, matrix, compression=compression)
        else:
            if format == "dense":
                group = self._file.create_dataset(
                    name=dest,
                    shape=matrix_shape,
                    dtype=data_dtype,
                    compression=compression,
                )
                # https://anndata.readthedocs.io/en/latest/fileformat-prose.html#dense-arrays-specification-v0-2-0
                group.attrs["encoding-type"] = "array"
                group.attrs["encoding-version"] = "0.2.0"
            elif format in [
                "csr",
                "csc",
            ]:  # Based on https://github.com/appier/h5sparse/blob/master/h5sparse/h5sparse.py
                if data_dtype is None:
                    data_dtype = np.float64
                if matrix_shape is None:
                    matrix_shape = (0, 0)
                sparse_class = ss.csr_matrix if format == "csr" else ss.csc_matrix
                data = sparse_class(matrix_shape, dtype=data_dtype)
                self._write_elem(dest, data, compression=compression)
            else:
                raise NotImplementedError(
                    f"Format must  be 'dense', 'csr' or 'csc' but {format} given!"
                )


class RawLayer(BaseLayerMatrixAndDf):
    def __init__(self, h5_file: h5py.File):
        super().__init__(h5_file, path_to_content="/raw/")
        self._var: CapAnnDataDF = None

    @property
    def var(self) -> CapAnnDataDF:
        if self._var is None:
            self._var = self._lazy_df_load("var")
        return self._var

    @var.setter
    def var(self, cap_df: CapAnnDataDF) -> None:
        self._validate_cap_df(cap_df, axis=1)
        self._var = cap_df

    def read_var(self, columns: List[str] = None, reset: bool = False) -> None:
        df = self._read_df(key="var", columns=columns)
        if self.var.empty or reset:
            self._var = df
        else:
            for col in df.columns:
                self._var[col] = df[col]


class CapAnnData(BaseLayerMatrixAndDf):
    def __init__(self, h5_file: h5py.File) -> None:
        super().__init__(h5_file, path_to_content="/")
        self._file: h5py.File = h5_file
        self._obs: CapAnnDataDF = None
        self._var: CapAnnDataDF = None
        self._X: X_NOTATION = None
        self._obsm: CapAnnDataDict = None
        self._varm: CapAnnDataDict = None
        self._layers: CapAnnDataDict = None
        self._uns: CapAnnDataDict = None
        self._obsp: CapAnnDataDict = None
        self._varp: CapAnnDataDict = None
        self._raw: RawLayer = None
        self._shape: Tuple[int, int] = None

    @property
    def obs(self) -> CapAnnDataDF:
        if self._obs is None:
            self._obs = self._lazy_df_load("obs")
        return self._obs

    @obs.setter
    def obs(self, cap_df: CapAnnDataDF) -> None:
        self._validate_cap_df(cap_df, axis=0)
        self._obs = cap_df

    @property
    def var(self) -> CapAnnDataDF:
        if self._var is None:
            self._var = self._lazy_df_load("var")
        return self._var

    @var.setter
    def var(self, cap_df: CapAnnDataDF) -> None:
        self._validate_cap_df(cap_df, axis=1)
        self._var = cap_df

    @property
    def raw(self) -> RawLayer:
        if self._raw is None:
            if "raw" not in self._file.keys():
                logger.debug("Can't read raw.var since raw layer doesn't exist!")
                return

            if len(self._file["raw"].keys()) == 0:
                logger.debug("The raw layer is empty!")
                return

            self._raw = RawLayer(self._file)
        return self._raw

    @property
    def uns(self) -> CapAnnDataDict[str, Any]:
        if self._uns is None:
            self._uns = CapAnnDataDict(
                {k: NotLinkedObject for k in self._file["uns"].keys()}
            )
        return self._uns

    @property
    def layers(self) -> CapAnnDataDict[str, X_NOTATION]:
        if self._layers is None:
            self._link_layers()
        return self._layers

    @property
    def obsm(self) -> CapAnnDataDict[str, X_NOTATION]:
        if self._obsm is None:
            self._link_obsm()
        return self._obsm

    @property
    def varm(self) -> CapAnnDataDict[str, X_NOTATION]:
        if self._varm is None:
            self._link_varm()
        return self._varm

    @property
    def obsp(self) -> CapAnnDataDict[str, X_NOTATION]:
        if self._obsp is None:
            self._link_obsp()
        return self._obsp

    @property
    def varp(self) -> CapAnnDataDict[str, X_NOTATION]:
        if self._varp is None:
            self._link_varp()
        return self._varp

    def read_obs(self, columns: List[str] = None, reset: bool = False) -> None:
        df = self._read_df("obs", columns=columns)
        if self.obs.empty or reset:
            self._obs = df
        else:
            for col in df.columns:
                self._obs[col] = df[col]

    def read_var(self, columns: List[str] = None, reset: bool = False) -> None:
        df = self._read_df("var", columns=columns)
        if self.var.empty or reset:
            self._var = df
        else:
            for col in df.columns:
                self._var[col] = df[col]

    def read_uns(self, keys: List[str] = None) -> None:
        if keys is None:
            keys = list(self.uns.keys())

        for key in keys:
            existing_keys = self.uns.keys()
            if key in existing_keys:
                source = self._file[f"uns/{key}"]
                self.uns[key] = read_elem(source)

    def _link_layers(self) -> None:
        if self._layers is None:
            self._layers = CapAnnDataDict()
        if "layers" in self._file.keys():
            self._link_array_mapping(cap_dict=self._layers, key="layers")

    def _link_obsm(self) -> None:
        key = "obsm"
        if self._obsm is None:
            self._obsm = CapAnnDataDict()
        if key in self._file.keys():
            self._link_array_mapping(cap_dict=self._obsm, key=key)

    def _link_varm(self) -> None:
        key = "varm"
        if self._varm is None:
            self._varm = CapAnnDataDict()
        if key in self._file.keys():
            self._link_array_mapping(cap_dict=self._varm, key=key)

    def _link_obsp(self):
        key = "obsp"
        if self._obsp is None:
            self._obsp = CapAnnDataDict()

        if key in self._file.keys():
            self._link_array_mapping(cap_dict=self._obsp, key=key)

    def _link_varp(self):
        key = "varp"
        if self._varp is None:
            self._varp = CapAnnDataDict()

        if key in self._file.keys():
            self._link_array_mapping(cap_dict=self._varp, key=key)

    def obsm_keys(self) -> List[str]:
        return list(self.obsm.keys())

    def obs_keys(self) -> List[str]:
        return self.obs.column_order_array().tolist()

    def var_keys(self) -> List[str]:
        return self.var.column_order_array().tolist()
    
    def field_to_entity(self, key):
        if key == "obs":
            return self.obs
        elif key == "var":
            return self.var
        elif key == "raw.var":
            return self.raw.var if self.raw is not None else None
        elif key == "uns":
            return self.uns
        elif key == "layers":
            return self.layers
        elif key == "obsm":
            return self.obsm
        elif key == "varm":
            return self.varm
        elif key == "obsp":
            return self.obsp
        elif key == "varp":
            return self.varp
        else:
            raise KeyError(
                f"The field {key} is not supported! The list of supported fields are equal to {FIELDS_SUPPORTED_TO_OVERWRITE} "
                f"attributes of the CapAnnData class."
            )

    def overwrite(self, fields: List[str] = None, compression: str = "lzf") -> None:
        if fields is None:
            fields = FIELDS_SUPPORTED_TO_OVERWRITE

        for key in ["obs", "var", "raw.var"]:
            if key in fields:
                entity: CapAnnDataDF = self.field_to_entity(key)
                if entity is None:
                    continue

                key = key.replace(".", "/") if key == "raw.var" else key

                for col in entity.columns:
                    self._write_elem(
                        f"{key}/{col}", entity[col].values, compression=compression
                    )

                column_order = entity.column_order_array()
                if (
                    column_order.size == 0
                ):  # Refs https://github.com/cellannotation/cap-anndata/issues/6
                    column_order = np.array([], dtype=np.float64)

                # Index update
                index_name = entity.index.name
                if not index_name:
                    index_name = "_index"
                self._file[key].attrs["_index"] = index_name
                index_col = self._read_attr(self._file[key], "_index")
                self._write_elem(
                    f"{key}/{index_col}", entity.index.to_numpy(), compression=compression
                )

                self._file[key].attrs["column-order"] = column_order

        if "uns" in fields:
            for key in self.uns.keys():
                if self.uns[key] is not NotLinkedObject:
                    dest = f"uns/{key}"
                    self._write_elem(dest, self.uns[key], compression=compression)
            for key in self.uns.keys_to_remove:
                del self._file[f"uns/{key}"]

        for field in ["layers", "obsm", "varm", "obsp", "varp"]:
            if field in fields:
                for key in self.field_to_entity(field).keys_to_remove:
                    del self._file[f"{field}/{key}"]

    def create_layer(
        self,
        name: str,
        matrix: Union[np.ndarray, ss.csr_matrix, ss.csc_matrix, None] = None,
        matrix_shape: Union[tuple[int, int], None] = None,
        data_dtype: Union[np.dtype, None] = None,
        format: Union[str, None] = None,
        compression: str = "lzf",
    ) -> None:
        """
        The empty layer will be created in the case of `matrix` is None.
        """
        self._create_new_matrix_in_field(
            field="layers",
            name=name,
            matrix=matrix,
            matrix_shape=matrix_shape,
            data_dtype=data_dtype,
            format=format,
            compression=compression,
        )
        self._link_layers()

    def create_obsm(
        self,
        name: str,
        matrix: Union[np.ndarray, ss.csr_matrix, ss.csc_matrix, None] = None,
        matrix_shape: Union[tuple[int, int], None] = None,
        data_dtype: Union[np.dtype, None] = None,
        format: Union[str, None] = None,
        compression: str = "lzf",
    ) -> None:
        self._create_new_matrix_in_field(
            field="obsm",
            name=name,
            matrix=matrix,
            matrix_shape=matrix_shape,
            data_dtype=data_dtype,
            format=format,
            compression=compression,
        )
        self._link_obsm()

    def create_varm(
        self,
        name: str,
        matrix: Union[np.ndarray, ss.csr_matrix, ss.csc_matrix, None] = None,
        matrix_shape: Union[tuple[int, int], None] = None,
        data_dtype: Union[np.dtype, None] = None,
        format: Union[str, None] = None,
        compression: str = "lzf",
    ) -> None:
        self._create_new_matrix_in_field(
            field="varm",
            name=name,
            matrix=matrix,
            matrix_shape=matrix_shape,
            data_dtype=data_dtype,
            format=format,
            compression=compression,
        )
        self._link_varm()

    def create_obsp(
        self,
        name: str,
        matrix: Union[np.ndarray, ss.csr_matrix, ss.csc_matrix, None] = None,
        matrix_shape: Union[tuple[int, int], None] = None,
        data_dtype: Union[np.dtype, None] = None,
        format: Union[str, None] = None,
        compression: str = "lzf",
    ) -> None:
        self._create_new_matrix_in_field(
            field="obsp",
            name=name,
            matrix=matrix,
            matrix_shape=matrix_shape,
            data_dtype=data_dtype,
            format=format,
            compression=compression,
        )
        self._link_obsp()

    def create_varp(
        self,
        name: str,
        matrix: Union[np.ndarray, ss.csr_matrix, ss.csc_matrix, None] = None,
        matrix_shape: Union[tuple[int, int], None] = None,
        data_dtype: Union[np.dtype, None] = None,
        format: Union[str, None] = None,
        compression: str = "lzf",
    ) -> None:

        self._create_new_matrix_in_field(
            field="varp",
            name=name,
            matrix=matrix,
            matrix_shape=matrix_shape,
            data_dtype=data_dtype,
            format=format,
            compression=compression,
        )
        self._link_varp()

    def _create_new_matrix_in_field(self, field, name, **kwargs):
        """**kwargs: matrix, matrix_shape, data_dtype, format, compression"""
        dest = f"{field}/{name}"
        field_entity = getattr(self, field)
        if name in field_entity.keys():
            raise ValueError(
                f"Please explicitly remove the existing '{name}' entity from {field} "
                f"before creating a new one!"
            )
        if field not in self._file.keys():
            self._file.create_group(field)
        self._create_new_matrix(dest=dest, **kwargs)

    def remove_layer(self, name: str) -> None:
        del self._file[f"layers/{name}"]
        self._link_layers()

    def remove_obsp(self, name: str) -> None:
        del self._file[f"obsp/{name}"]
        self._link_obsp()

    def remove_varp(self, name: str) -> None:
        del self._file[f"varp/{name}"]
        self._link_varp()

    def remove_obsm(self, name: str) -> None:
        del self._file[f"obsm/{name}"]
        self._link_obsm()

    def remove_varm(self, name: str) -> None:
        del self._file[f"varm/{name}"]
        self._link_varm()

    def create_repr(self) -> str:
        indent = " " * 4
        s = f"CapAnnData object"
        s += f"\n{indent}File: {self._file}"
        s += f"\n{indent}X shape: {self.shape}"
        s += f"\n{indent}Has raw X: {self.raw is not None}"
        for field in ["obs", "obsm", "var", "uns", "layers"]:
            if field in self._file:
                in_memory = set()
                if field in ["obs", "var", "uns"]:
                    attr = getattr(self, field)
                    if attr is not None:
                        in_memory = set(attr.keys())
                keys = list(self._file[field].keys())
                keys = [k for k in keys if k != "_index"]
                keys = [(k if k not in in_memory else f"{k}*") for k in keys]
                keys_str = str(keys).replace("*'", "'*")
                s += f"\n{indent}{field}: {keys_str}"
        s += f"\n{indent}Note: fields marked with * are in-memory objects."
        return s

    def __repr__(self) -> str:
        return self.create_repr()

    def __str__(self) -> str:
        return self.create_repr()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self._file is not None:
            self._file.close()
        logger.debug("CapAnnData closed!")



from setuptools import setup, find_packages

setup(
    name='cap_anndata',
    version='0.5.0',
    author='R. Mukhin, A. Isaev',
    author_email='roman@ebookapplications.com',
    packages=find_packages(exclude=["test"]),
    description='Partial read/write of AnnData (h5ad) files for low-memory operations with large datasets.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/cellannotation/cap-anndata',
    project_urls={
        "Bug Tracker": "https://github.com/cellannotation/cap-anndata/issues",
        "Changelog": "https://github.com/cellannotation/cap-anndata/blob/main/CHANGELOG.md",
        "Documentation": "https://github.com/cellannotation/cap-anndata/blob/main/HOWTO.md",
    },
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    install_requires=[
        "numpy>=1.23.5",
        "pandas>=2.2.0",
        "anndata>=0.10.0",
    ],
    extras_require={"dev": [
        "pytest>=8.0.0",
        "setuptools~=69.1.1"
    ]}
)



import logging
import h5py
import warnings

from cap_anndata import CapAnnData


logger = logging.getLogger(__name__)


def read_h5ad(file_path: str, edit: bool = False):
    """
    This is the main read method for CapAnnData.
    Must be used in 'with' context.
    """
    mode = "r+" if edit else "r"
    logger.debug(f"Read file {file_path} mode={mode} in context...")

    try:
        file = h5py.File(file_path, mode)
        cap_adata = CapAnnData(file)
        logger.debug(f"Successfully read anndata file path {file_path}")
        return cap_adata

    except Exception as error:
        logger.error(
            f"Error during read anndata file at path: {file_path}, error = {error}!"
        )
        raise error


def deprecated(message):
    def deprecated_decorator(func):
        def deprecated_func(*args, **kwargs):
            warnings.warn(
                "{} is a deprecated function. {}".format(func.__name__, message),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return func(*args, **kwargs)

        return deprecated_func

    return deprecated_decorator


# TODO: remove deprecated function
@deprecated(
    "It will be removed in the next version of package. Please replace it with `read_h5ad`."
)
def read_directly(file_path: str, edit: bool = False) -> CapAnnData:
    """
    Must be used only in specific cases.
    User is responsible to close the h5py file when the work with CapAnnData instance done.
    """
    return read_h5ad(file_path, edit)
